# From:
# https://github.com/pallets/click/issues/2676

# The above issue is related to numerical negative arguments. I've adapted that
# solution to allow negative numbers in options with optional values

from gettext import ngettext
from click.parser import OptionParser, _flag_needs_value, BadOptionUsage


def _is_number(arg: str) -> bool:
    try:
        _ = int(arg, 0)
    except ValueError:
        try:
            _ = float(arg)
        except ValueError:
            return False
    return True


# Allow negative numbers in options with optional values
#
# Based on click/parser.py v8.1.8
#
def _get_value_from_state(
    self, option_name, option, state
):
    nargs = option.nargs

    if len(state.rargs) < nargs:
        if option.obj._flag_needs_value:
            # Option allows omitting the value.
            value = _flag_needs_value
        else:
            raise BadOptionUsage(
                option_name,
                ngettext(
                    "Option {name!r} requires an argument.",
                    "Option {name!r} requires {nargs} arguments.",
                    nargs,
                ).format(name=option_name, nargs=nargs),
            )
    elif nargs == 1:
        next_rarg = state.rargs[0]

        if (
            option.obj._flag_needs_value
            and isinstance(next_rarg, str)
            and next_rarg[:1] in self._opt_prefixes
            and len(next_rarg) > 1
            and not _is_number(next_rarg)
        ):
            # The next arg looks like the start of an option, don't
            # use it as the value if omitting the value is allowed.
            value = _flag_needs_value
        else:
            value = state.rargs.pop(0)
    else:
        value = tuple(state.rargs[:nargs])
        del state.rargs[:nargs]

    return value


# Monkey-patch the OptionParser class
#
OptionParser._get_value_from_state = _get_value_from_state



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
caldavctl uses a simple format to create new objects. this format is parsed and
converted into iCalendar objects.
'''

from abc import ABC, abstractmethod
from datetime import datetime
from functools import reduce
from operator import and_, xor
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError


class ValueParser(ABC):
    def __init__(self, txt):
        '''
        Arguments:

        txt - raw data
        '''
        self.txt = txt
        self.tokens = None
        self.value = None  # return value

    @abstractmethod
    def tokenize(self):
        ...

    @abstractmethod
    def validate(self):
        ...

    def parse(self):
        self.tokenize()
        self.validate()
        return self.value


class ParserError(Exception):
    pass


class TimezoneParser(ValueParser):
    def tokenize(self):
        self.tokens = self.txt

    def validate(self):
        try:
            self.value = ZoneInfo(self.tokens)
        except (ZoneInfoNotFoundError, ValueError):
            raise ParserError(f'invalid timezone: "{self.tokens}"')


class DatetimeParser(ValueParser):
    def tokenize(self):
        self.tokens = self.txt

    def validate(self):
        try:
            self.value = datetime.fromisoformat(self.tokens)
        except ValueError:
            raise ParserError(f'Invalid date format for: "{self.tokens}"')


def integer_parser(min=None, max=None):

    class IntegerParser(ValueParser):
        def tokenize(self):
            self.tokens = self.txt

        def validate(self):
            try:
                value = int(self.tokens)
            except ValueError as msg:
                raise ParserError(msg)
            if min and value < min:
                raise ParserError('the value is less than the minimum value.')
            if max and value > max:
                raise ParserError('the value is greater than the maximum value.')
            self.value = value

    return IntegerParser


def integer_list_parser(min=None, max=None, exception=[], sep=','):

    class IntegerListParser(ValueParser):
        def tokenize(self):
            self.tokens = self.txt.split(sep)

        def validate(self):
            try:
                values = [int(vl) for vl in self.tokens]
            except ValueError as msg:
                raise ParserError(msg)
            for value in values:
                if min and value < min:
                    raise ParserError(f'the value is less than the minimum value ({min}).')
                if max and value > max:
                    raise ParserError(f'the value is greater than the maximum value ({max}).')
                if value in exception:
                    raise ParserError(f'forbiden value {value}.')
            self.value = values

    return IntegerListParser


def choices_parser(choices, sep=','):
    choices = [el.upper() for el in choices]

    class EnumParser(ValueParser):
        def tokenize(self):
            self.tokens = self.txt

        def validate(self):
            if self.tokens.upper() not in choices:
                raise ParserError(f'Element {self.tokens} not in allowed values: {', '.join(choices)}')
            self.value = self.tokens

    return EnumParser


def string_parser(max_len=None):

    class StringParser(ValueParser):
        def tokenize(self):
            self.tokens = self.txt

        def validate(self):
            if max_len and len(self.tokens) > max_len:
                raise ParserError(f'max string lenght is {max_len}.')
            self.value = self.tokens

    return StringParser


def list_parser(max_len=None, sep=','):

    class ListParser(ValueParser):
        def tokenize(self):
            if self.txt:
                self.tokens = [el.strip() for el in self.txt.split(sep)]
            else:
                self.tokens = []

        def validate(self):
            if max_len and len(self.tokens) > max_len:
                raise ParserError(f'max list lenght is {max_len}.')
            self.value = self.tokens

    return ListParser


def flaten(lst):
    for el in lst:
        if isinstance(el, list):
            yield flaten(el)
        else:
            yield el


class ObjectParser(ABC):
    def __init__(self, obj, lexicon, mandatory=[], exclusive=[]):
        self.object = obj
        self.lexicon = lexicon
        self.mandatory = mandatory
        self.optional = set(lexicon.keys()) - set(flaten(mandatory))
        self.exclusive = exclusive

    @abstractmethod
    def tokenize(self):
        ...

    def check_mandatory(self, result):
        # Check mandatory fields
        for key in self.mandatory:
            if isinstance(key, list):
                if reduce(and_, [subkey in result for subkey in key]):
                    raise ParserError(f'Only one of these keys can be present: {', '.join(key)}')
                elif not reduce(xor, [subkey in result for subkey in key]):
                    raise ParserError(f'One of these keys must be present: {', '.join(key)}')
            else:
                if key not in result:
                    raise ParserError(f'Missing mandatory key: "{key}"')

    def check_exclusive(self, result):
        for exclusive_key in self.exclusive:
            if reduce(and_, [subkey in result for subkey in exclusive_key]):
                raise ParserError(f'Only one of these keys can be present: {', '.join(exclusive_key)}')

    def drop_optional(self, result):
        '''
        Remove empty optional keys from results
        '''
        for key in self.optional:
            if key in result and not result[key]:
                del result[key]

    def parse(self, tokens):
        result = {}
        for key, value in tokens:
            key = key.lower()  # The keys are case insensitive
            if key not in self.lexicon:
                raise ParserError(f'unknown key: "{key}".')

            value_type = self.lexicon[key]

            try:
                if value:
                    parsed_value = value_type(value).parse()
                else:
                    parsed_value = value
            except ParserError as msg:
                raise ParserError(f'Error evaluating key "{key.upper()}": {msg}')
            if key in result:
                # We have a key that can show up more than one time
                if not isinstance(result[key], list):
                    result[key] = [result[key]]
                result[key].append(parsed_value)
            else:
                # Single value key
                result[key] = parsed_value

        return result

    def run(self):
        tokens = self.tokenize()
        result = self.parse(tokens)
        self.check_mandatory(result)
        self.check_exclusive(result)
        self.drop_optional(result)
        return result


class EventParser(ObjectParser):

    def tokenize(self):
        # States:
        KEY = 0
        VALUE = 1

        data = self.object
        tokens = []
        pos = 0
        lenght = len(data)
        state = KEY
        while pos < lenght:
            if data[pos] in ' \t':  # Ignore spaces
                pos += 1
                continue
            if data[pos] == '#':  # Ignore comments
                pos = data.find('\n', pos) + 1
                if pos == 0:  # End of file
                    pos = lenght
                continue

            if state == KEY:  # Find the key
                if data[pos] in ';\n':  # New lines or semicolons before the key
                    pos += 1
                    continue
                end = data.find(':', pos)
                if end == -1:
                    raise ParserError('Could not find key/value pair.')
                key = data[pos:end].strip()
                pos = end + 1
                state = VALUE
                continue

            if state == VALUE:  # Find the value
                if pos == lenght or data[pos] in ';\n':  # End of file or empty value
                    pos += 1
                    tokens.append([key, ''])
                    state = KEY
                    continue
                if data[pos:pos + 2] == '[[':  # Multi line string
                    end = data.find(']]', pos)
                    if end == -1:  # No closing
                        raise ParserError('Unclosed double bracket')
                    value = data[pos + 2:end].strip()
                    tokens.append([key, value])
                    pos = end + 2
                    state = KEY
                    continue
                else:
                    # Single line string
                    semicolon = data.find(';', pos)
                    newline = data.find('\n', pos)
                    comment = data.find('#', pos)

                    if semicolon != -1 and newline != -1 and newline < semicolon:
                        end = newline
                    elif semicolon != -1 and newline != -1 and semicolon < newline:
                        end = semicolon
                    elif semicolon == -1 and newline == -1:  # end of file
                        end = lenght
                    elif semicolon == -1 and newline != -1:  # End of line no semicolons
                        end = newline
                    elif semicolon != -1 and newline == -1:  # Last line with semicolons
                        end = semicolon
                    else:  # Just to make sure I got all possible combinations
                        raise ParserError('Internal tokenizer error')

                    # Handle comments
                    if comment != -1 and comment < end:
                        value = data[pos:comment].strip()
                        end = comment - 1
                    else:
                        value = data[pos:end].strip()

                    tokens.append([key, value])
                    pos = end + 1
                    state = KEY
                    continue
            pos += 1

        if state == VALUE:
            tokens.append([key, ''])
        return tokens


class RRuleParser(ObjectParser):

    def tokenize(self):
        data = self.object
        return [
            [tok.upper().strip() for tok in keyvalue.split('=', maxsplit=1)]
            for keyvalue in data.split(';')
            if keyvalue.strip()
        ]



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
from datetime import datetime, timezone
from zoneinfo import ZoneInfo
from caldavctl.object_parser import EventParser, ParserError
from caldavctl.event_builder import parse_event, check_if_naive, EVENT_LEXICON, MANDATORY_KEYS


sample_event_data_i = """
        SUMMARY: This is a description
        LOCATION: [[Timbuktu]] # With a comment
        CATEGORIES: Birthday, Shopping, Training
        TIMEZONE: Europe/Lisbon
        DTSTART: 2024-01-20 09:00
        DTEND: 2024-01-20 09:45
        DESCRIPTION: [[
        Once upon a time,
        There was a little lamb
        ]]
        """

expected_tokens_i = [
    ['SUMMARY', 'This is a description'],
    ['LOCATION', 'Timbuktu'],
    ['CATEGORIES', 'Birthday, Shopping, Training'],
    ['TIMEZONE', "Europe/Lisbon"],
    ['DTSTART', '2024-01-20 09:00'],
    ['DTEND', '2024-01-20 09:45'],
    ['DESCRIPTION', 'Once upon a time,\n        There was a little lamb']
]

sample_event_data_ii = '''# Mandatory
SUMMARY: Example summary
DTSTART: 2024-12-30 09:00
DTEND: 2024-12-30 09:45

# Optional
LOCATION:
CATEGORIES:
TIMEZONE: Europe/Lisbon
DESCRIPTION: [[ ]]

# NOTES:
#
#   * Date and time:
#       * The dates must be in iso format, for instance: 2024-12-29 13:45;
#       * The timezone used if the one defined by default or the one defined in
#         "TIMEZONE";
#   * Categories: the categories are a comma separeted list;
#   * Description: The description can be multi line, just make sure it's
#     delimited by [[ ]].'''


@pytest.fixture
def default_timezone():
    return ZoneInfo('UTC')


tokenize_test_data = (
    (sample_event_data_i, expected_tokens_i),
    # Multiple keys in one line
    ('KEY1:VALUE1; KEY2:VALUE2;', [['KEY1', 'VALUE1'], ['KEY2', 'VALUE2']]),
    ('KEY1:VALUE1; KEY2:VALUE2\nKEY3:VALUE3;', [['KEY1', 'VALUE1'], ['KEY2', 'VALUE2'],
                                                ['KEY3', 'VALUE3']]),

    # Empty values
    ('KEY1:', [['KEY1', '']]),
    ('KEY1:\nKEY2:', [['KEY1', ''], ['KEY2', '']]),
    ('KEY1:   ', [['KEY1', '']]),  # Whitespace that should be trimmed
    ('KEY1:; KEY2:value', [['KEY1', ''], ['KEY2', 'value']]),

    # Comments handling
    ('KEY: value # comment', [['KEY', 'value']]),
    ('# Just a comment\nKEY: value', [['KEY', 'value']]),
    ('KEY: value with # hash', [['KEY', 'value with']]),
    ('KEY: value # comment\nKEY2: value2', [['KEY', 'value'], ['KEY2', 'value2']]),

    # Whitespace handling
    ('  KEY:  value  ', [['KEY', 'value']]),
    ('KEY:value', [['KEY', 'value']]),
    (' KEY : value ', [['KEY', 'value']]),

    # Multiline values with brackets
    ('''KEY: [[
    multiline
    value
    ]]''', [['KEY', 'multiline\n    value']]),

    # Nested brackets
    ('KEY: [[ value with [nested] brackets ]]', [['KEY', 'value with [nested] brackets']]),

    # Mixed delimiters
    ('KEY1: value1; KEY2:value2', [['KEY1', 'value1'], ['KEY2', 'value2']]),
    ('KEY1: value1\nKEY2:value2', [['KEY1', 'value1'], ['KEY2', 'value2']]),

    # Case sensitivity
    ('key: value', [['key', 'value']]),
    ('KEY: value', [['KEY', 'value']]),

    # Special characters in values
    ('KEY: value:with:colons', [['KEY', 'value:with:colons']]),

    # Empty keys (should probably be rejected by the parser but good to test)
    (': value without key', [['', 'value without key']]),


    # Mixed line endings
    ('KEY1: value1\r\nKEY2: value2', [['KEY1', 'value1'], ['KEY2', 'value2']]),

    # Multiline with comments
    ('''KEY: [[
    multiline # not a comment
    value
    ]] # this is a comment''', [['KEY', 'multiline # not a comment\n    value']]),

    # Multiple multiline values
    ('''KEY1: [[
    first multiline
    ]]
    KEY2: [[
    second multiline
    ]]''', [['KEY1', 'first multiline'], ['KEY2', 'second multiline']]),

    # TODO: Should escaped characters be supported?
    # Escaped characters
    ('KEY: value with \\"quotes\\"', [['KEY', 'value with \\"quotes\\"']]),
    ('KEY: value with \\n newline', [['KEY', 'value with \\n newline']]),

    # Multiple keys with same name (should preserve all)
    ('KEY: value1\nKEY: value2', [['KEY', 'value1'], ['KEY', 'value2']]),

    # Keys with spaces or unusual characters (if supported)
    ('KEY WITH SPACES: value', [['KEY WITH SPACES', 'value']]),
    ('KEY-WITH-DASHES: value', [['KEY-WITH-DASHES', 'value']]),
    ('KEY_WITH_UNDERSCORES: value', [['KEY_WITH_UNDERSCORES', 'value']]),

    # TODO: should this be supported?
    # # Mixed line continuation
    # ('''KEY: value that \
    # continues on next line''', [['KEY', 'value that continues on next line']]),

    # Completely empty input
    ('', []),

    # Just whitespace
    ('   \n  \t  ', []),

    # Just comments
    ('# comment1\n# comment2', []),

    # Comments
    ('KEY1: value 1 # ; KEY2: value2', [['KEY1', 'value 1']])

)


@pytest.mark.parametrize('sample, result', tokenize_test_data)
def test_tokenize_data(sample, result):
    print(f'\nTesting: {sample}')
    parser = EventParser(sample, EVENT_LEXICON, MANDATORY_KEYS)
    tokens = parser.tokenize()
    print('Computed tokens:', tokens)
    assert tokens == result


tokenizer_error_test_data = (
    # Unclosed multiline blocks
    ('''KEY: [[
    unclosed multiline
    ''', 'Unclosed double bracket'),

    ('KEY: value;with;semicolons', 'Could not find key/value pair'),

)


@pytest.mark.parametrize('sample, error_message', tokenizer_error_test_data)
def test_tokenizer_errors(sample, error_message):
    print(f'\nTesting: {sample}')
    parser = EventParser(sample, EVENT_LEXICON, MANDATORY_KEYS)
    with pytest.raises(ParserError, match=error_message):
        tokens = parser.tokenize()
        print('Computed tokens:', tokens)


def test_parse_i():
    result = EventParser(sample_event_data_i, EVENT_LEXICON, MANDATORY_KEYS).run()

    expected_result = {
        'summary': 'This is a description',
        'location': 'Timbuktu',
        'categories': ['Birthday', 'Shopping', 'Training'],
        'timezone': ZoneInfo('Europe/Lisbon'),
        'dtstart': datetime(2024, 1, 20, 9, 0),
        'dtend': datetime(2024, 1, 20, 9, 45),
        'description': 'Once upon a time,\n        There was a little lamb'
    }

    assert result['summary'] == expected_result['summary']
    assert result['location'] == expected_result['location']
    assert result['categories'] == expected_result['categories']
    assert result['description'] == expected_result['description']
    assert result['dtstart'] == expected_result['dtstart']
    assert result['dtend'] == expected_result['dtend']
    assert result['timezone'] == expected_result['timezone']


def test_parse_ii():
    result = EventParser(sample_event_data_ii, EVENT_LEXICON, MANDATORY_KEYS).run()

    expected_result = {
        'summary': 'Example summary',
        'location': '',
        'categories': [],
        'timezone': ZoneInfo('Europe/Lisbon'),
        'dtstart': datetime(2024, 12, 30, 9, 0),
        'dtend': datetime(2024, 12, 30, 9, 45),
        'description': ''
    }

    assert result['summary'] == expected_result['summary']
    assert 'location' not in result
    assert 'categories' not in result
    assert 'description' not in result
    assert result['dtstart'] == expected_result['dtstart']
    assert result['dtend'] == expected_result['dtend']
    assert result['timezone'] == expected_result['timezone']


def test_check_if_naive():
    naive_date = datetime(2024, 1, 20, 9, 0)
    tz = ZoneInfo('Europe/Lisbon')
    localized_date = check_if_naive(naive_date, tz)

    assert f'{localized_date.tzinfo}' == 'Europe/Lisbon'
    assert localized_date.utcoffset() == tz.utcoffset(naive_date)


def test_parse_event_with_naive_dates(default_timezone):
    event_with_naive_dates = (
        """
        SUMMARY: Test Event
        LOCATION: Timbuktu
        TIMEZONE: Europe/Lisbon
        DTSTART: 2024-01-20 09:00
        DTEND: 2024-01-20 09:45Z  # UTC date
        """
    )

    result = parse_event(event_with_naive_dates, default_timezone)
    tz = ZoneInfo('Europe/Lisbon')

    assert result['dtstart'].tzinfo == tz
    assert result['dtend'].tzinfo == timezone.utc


def test_unknown_key():
    event_with_unknown_key = (
        """
        SUMMARY: Test Event
        UNKNOWN_KEY: This should fail
        """
    )
    parser = EventParser(event_with_unknown_key, EVENT_LEXICON, MANDATORY_KEYS)

    with pytest.raises(ParserError, match='unknown key:'):
        parser.run()


def test_invalid_date():
    event_with_invalid_date = (
        """
        SUMMARY: Test Event
        DTSTART: invalid-date
        """
    )

    parser = EventParser(event_with_invalid_date, EVENT_LEXICON, MANDATORY_KEYS)

    with pytest.raises(ParserError, match='Invalid date format'):
        parser.run()


def test_event_no_timezone_i():
    sample_event_no_timezone = '''# Mandatory
    SUMMARY: Example summary
    DTSTART: 2024-12-30 09:00
    DTEND: 2024-12-30 09:45'''

    result = parse_event(sample_event_no_timezone, ZoneInfo('Europe/Lisbon'))

    assert f'{result['dtstart'].tzinfo}' == 'Europe/Lisbon'
    assert f'{result['dtend'].tzinfo}' == 'Europe/Lisbon'


def test_event_no_timezone_ii():
    sample_event_no_timezone = '''# Mandatory
    SUMMARY: Example summary
    DTSTART: 2024-12-30 09:00
    DTEND: 2024-12-30 09:45

    # Optional
    LOCATION:
    CATEGORIES:
    TIMEZONE:
    PRIORITY: 5
    ALARM: P0D
    RRULE:
    DESCRIPTION: '''

    result = parse_event(sample_event_no_timezone, ZoneInfo('Europe/Lisbon'))

    assert f'{result['dtstart'].tzinfo}' == 'Europe/Lisbon'
    assert f'{result['dtend'].tzinfo}' == 'Europe/Lisbon'


def test_event_empty_event():
    sample_event_empty = ''

    with pytest.raises(ParserError,
                       match='Missing mandatory key: "dtstart"'):
        parse_event(sample_event_empty, ZoneInfo('Europe/Lisbon'))



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from datetime import datetime, timedelta

import click
from caldav.lib import error

import caldavctl.patch_click
from caldavctl import dav, get_name
from caldavctl.click_utils import create_compatibility_check
from caldavctl.event_builder import event_builder, parse_event
from caldavctl.object_parser import ParserError
from caldavctl.templates import render
from caldavctl.utils import (
    ESCAPE_SEQUENCES,
    calmonth,
    edit_text_with_editor,
    timedelta_to_duration,
    to_datetime)

# Configuration

EMPTY_EVENT = '''# Mandatory
DTSTART: %(start)s
DURATION: PT45M
# Use either DURATION or DTEND but not both
# DTEND: %(end)s

%(cal)s

# Recommended
SUMMARY:

# Optional
LOCATION:
CATEGORIES:
TIMEZONE:
PRIORITY: 5
ALARM: P0D
# RRULE: [[FREQ=DAILY;COUNT=3]]
DESCRIPTION: [[ ]]

# NOTES:
#
#   * Date and time:
#       * The dates must be in iso format, for instance: 2024-12-29 13:45;
#       * The timezone used is the one defined by default or the one defined in
#         the "TIMEZONE" key;
#       * The default time zone is "%(tz)s";
#       * You can use the command "caldavctl utils list-timezones" to list the
#         available time zones;
#       * The duration key has the format defined in:
#           https://www.rfc-editor.org/rfc/rfc5545#section-3.3.6
#       * The default duration is 45 minutes;
#   * Categories: the categories are a comma separated list;
#   * Priority: 1 - highest priority, 9 - lowest priority;
#   * Description: The description can be multi line, just make sure it's
#     delimited by [[ ]];
#   * Recurrence Rule (RRULE) allows scheduling recurring events. The syntaxe
#     used is defined in:
#     https://www.rfc-editor.org/rfc/rfc5545#section-3.3.10
#     The recurrence rule must be enclosed in [[ ]]
#   * Alarms:
#       * By default an alarm is defined the moment the event begins;
#       * The alarms are defined using:
#           ALARM: <trigger> [<description>]
#       * The trigger is defined according to:
#           https://www.rfc-editor.org/rfc/rfc5545#section-3.3.6
#       * For example, to trigger an alarm 30 minutes before the event start, use:
#           ALARM: -PT30M
#       * The alarm description is optional.
'''


# Commands:

exclusive_option = create_compatibility_check(
    ['today', 'week', 'dtstart', 'dtend', 'day'],  # Exclusive options
    [['dtstart', 'dtend']])  # Except these two that can be used together


@click.command('list', options_metavar='[options]')
@click.option('-t', '--today',
              is_flag=False, flag_value=0, default=None, type=int,
              callback=exclusive_option, help='Today\'s events', metavar='<offset>')
@click.option('-w', '--week',
              is_flag=False, flag_value=0, default=None, type=int,
              callback=exclusive_option, help='Week events', metavar='<offset>')
@click.option('-s', '--dtstart',
              callback=exclusive_option, help='Date range, start date', metavar='<yyyy-mm-dd>')
@click.option('-e', '--dtend',
              callback=exclusive_option, help='Date range, end date', metavar='<yyyy-mm-dd>')
@click.option('-d', '--day',
              callback=exclusive_option, help='Events on the day', metavar='<yyyy-mm-dd>')
@click.option('-sd', '--show-description',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s description.')
@click.option('-si', '--show-uid',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s UID.')
@click.option('-stz', '--show-timezone',
              is_flag=True, show_default=True, default=False,
              help='Show the event\'s time zones in the dates displayed.')
@click.option('-tf', '--template-file', default='event.txt',
              help='Template used to format the output', metavar='<file>')
@click.option('--json',
              is_flag=True, show_default=True, default=False,
              help='Output JSON instead of using a template')
@click.pass_obj
def list_events(context, today, week, dtstart, dtend, day, show_description, show_uid, show_timezone, template_file, json):
    '''
    List events from the server list

    The date range can be specified in several ways. The --week and --today
    options can take an integer offset, either positive or negative. For
    instace to see the events from last week we can use --week -1, to see the
    events for tomorrow use -t 1.

    caldavctl uses templates to display the list command output by default. If
    not specified otherwise  it uses the `event.txt` template. The used
    template can be specified using the -tf/--template-file option. First we
    check the current directory for the existence of the file. If it's not
    found a number of directories are checked for the template file. This
    command lists the template directories:

    ```
    caldavctl utils template-directories
    ```

    Currently we have the following templates:

    \b
        * event.txt - output with colors (default)
        * event-nocolor.txt
        * event.html
        * event-fancy.html

    If the --json flag is used, the list command will output JSON. This output
    does not go through the template system, and any template options will be
    ignored when using --json.
    '''
    tz = context['config'].tz()
    start_date = None
    end_date = None

    # Compute the start and end dates used on the search
    if today is not None:
        start_date = datetime.now().replace(hour=0, minute=0, second=0, tzinfo=tz)
        end_date = datetime.now().replace(hour=23, minute=59, second=59, tzinfo=tz)
        # Apply day offset
        start_date += timedelta(days=today)
        end_date += timedelta(days=today)
    elif week is not None:
        today_date = datetime.now().replace(hour=0, minute=0, second=0, tzinfo=tz)
        start_date = today_date - timedelta(days=today_date.weekday())
        end_date = start_date + timedelta(days=6)
        end_date = end_date.replace(hour=23, minute=59, second=59)
        # Apply week offset
        start_date += timedelta(weeks=week)
        end_date += timedelta(weeks=week)
    elif dtstart or dtend:
        if dtstart:
            start_date = to_datetime(dtstart, tz)
        else:
            start_date = datetime(1900, 1, 1, tzinfo=tz)
        if dtend:
            end_date = to_datetime(dtend, tz)
        else:
            end_date = datetime(5000, 1, 1, tzinfo=tz)
    elif day:
        try:
            dt = datetime.fromisoformat(day)
            start_date = dt.replace(hour=0, minute=0, second=0, tzinfo=tz)
            end_date = dt.replace(hour=23, minute=59, second=59, tzinfo=tz)
        except ValueError:
            raise click.UsageError(f'Invalid date "{day}"')
    else:
        raise click.UsageError('You have to specify a time frame, for example with -t or -w.')

    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()
    event_list = []
    with dav.caldav_calendar(server, calendar_id) as calendar:
        # Check if the calendar supports events:
        if 'VEVENT' not in calendar.get_supported_components():
            raise click.UsageError('This calendar does not support events.')

        # Get events in time range
        events = calendar.search(
            start=start_date,
            end=end_date,
            event=True,
            expand=True,
            sort_keys=['dtstart']
        )
        for event in events:
            ev = event.icalendar_component
            ev_start_date = ev.get('dtstart').dt
            if isinstance(ev_start_date, datetime):
                ev_start_date = ev_start_date.astimezone(tz)
            if 'dtend' in ev:
                ev_end_date = ev.get('dtend').dt
                if isinstance(ev_end_date, datetime):
                    ev_end_date = ev_end_date.astimezone(tz)
                duration = ev_end_date - ev_start_date
            else:
                duration = ev.get('duration').dt
                ev_end_date = ev_start_date + duration
            event_context = {
                'start_date': ev_start_date.strftime('%Y-%m-%d %H:%M') if isinstance(ev_start_date, datetime) else f'{ev_start_date} (day)',
                'start_date_tz': f'{ev_start_date.tzinfo}' if isinstance(ev_start_date, datetime) else None,
                'end_date': ev_end_date.strftime('%Y-%m-%d %H:%M') if isinstance(ev_end_date, datetime) else f'{ev_end_date} (day)',
                'end_date_tz': f'{ev_end_date.tzinfo}' if isinstance(ev_end_date, datetime) else None,
                'duration': f'{duration}',
                'duration_st': timedelta_to_duration(duration),
                'summary': ev.get('summary', ''),
                'description': ev.get('description', ''),
                'categories': [
                    cat if isinstance(cat, str) else cat.cats[0]
                    for cat in ev.get('categories', [])],
                'location': ev.get('location', ''),
                'uid': ev.get('uid')
            }
            event_list.append(event_context)

    if json:
        import json
        click.echo(json.dumps(event_list))
        return

    # Template output
    template_context = {
        'show_uid': show_uid,
        'show_description': show_description,
        'show_timezone': show_timezone,
        'event_list': event_list,
    }
    template_context.update(ESCAPE_SEQUENCES)

    txt = render(get_name(), template_file, template_context)
    click.echo(txt.strip())


@click.command('create', options_metavar='[options]')
@click.option('-f', '--file',
              type=click.File('r'), default=sys.stdin,
              help='Create event from this file', metavar='<file>')
@click.option('-e', '--edit', is_flag=True, default=False, help='Edit the event on a text editor?')
@click.option('--dry-run', is_flag=True, default=False, help='Show only the iCalendar generated')
@click.pass_obj
def create_event(context, file, edit, dry_run):
    '''
    Create new event

    The event data can be read from stdin or from a file. Optionally the event
    can be edited using the default $EDITOR. If the option to edit is enabled
    and no file is defined using -f/--file, then an empty event is opened.

    \b
    CalDAV CLI Event Configuration File Format
    ==========================================

    This format is used to define or edit calendar events via a simple
    key-value configuration file.

    \b
    Overview
    --------

    \b
    - Each line defines a field using the format: `KEY: VALUE`.
    - Lines starting with `#` are treated as comments and ignored.
    - Multi-line values (e.g., `DESCRIPTION`, `RRULE`) must be enclosed in `[[ ]]`.

    \b
    Sections
    --------

    \b
    Mandatory Fields
    ----------------

    \b
    DTSTART :
        Start date and time of the event.
        Format: `YYYY-MM-DD HH:MM`
        Example: `DTSTART: 2025-08-05 09:00`

    \b
    DURATION :
        Duration of the event based in the ISO 8601 format.
        See: https://www.rfc-editor.org/rfc/rfc5545#section-3.3.6
        Example: `DURATION: PT45M`

    \b
    DTEND :
        End date and time of the event.
        Format: `YYYY-MM-DD HH:MM`

    \b
        Notes
        -----
        Either `DURATION` or `DTEND` must be provided, but not both.

    \b
    Recommended Fields
    ------------------

    \b
    SUMMARY :
        Short summary or title of the event.
        Example: `SUMMARY: Team Meeting`

    \b
    Optional Fields
    ---------------

    \b
    LOCATION :
        Location where the event takes place.
        Example: `LOCATION: Conference Room A`

    \b
    CATEGORIES :
        Comma-separated list of categories/tags.
        Example: `CATEGORIES: Work,Team`

    \b
    TIMEZONE :
        Time zone in which to interpret the date/time values.
        Example: `TIMEZONE: Europe/Lisbon`

    \b
    PRIORITY :
        Priority of the event (1 = highest, 9 = lowest).
        Example: `PRIORITY: 5`

    \b
    ALARM :
        Defines when a reminder is triggered for the event.
        Format: `ALARM: <TRIGGER> [<DESCRIPTION>]`
        Examples:
            - `ALARM: P0D` (trigger at start)
            - `ALARM: -PT30M Reminder: Prepare for meeting` (trigger half an
              hour before)

    \b
    RRULE :
        Recurrence rule using RFC 5545 format.
        Must be enclosed in `[[ ]]`.
        See: https://www.rfc-editor.org/rfc/rfc5545#section-3.3.10
        Example: `RRULE: [[FREQ=DAILY;COUNT=3]]`

    \b
    DESCRIPTION :
        Optional multi-line description of the event.
        Must be enclosed in `[[ ]]`.
        Example:
            ```
            DESCRIPTION: [[
            Planning session with the engineering team.
            Agenda:
            - Sprint Review
            - Roadmap discussion
            ]]
            ```

    \b
    Comment Syntax
    --------------

    Lines beginning with `#` are treated as comments and ignored.

    \b
    Other Notes
    -----------

    \b
    - Date/Time values must be in ISO format: `YYYY-MM-DD HH:MM`
    - DURATION follows ISO 8601 duration syntax (e.g., `PT45M`, `P1DT2H`)
    - Use `caldavctl utils list-timezones` to see all available time zones
    - Default values:
        - DURATION: 45 minutes
        - ALARM: P0D (triggers at event start)
    - DESCRIPTION may use double square brackets `[[ ]]` to support multi-line
      or special content
    - RRULE must use double square brackets `[[ ]]` to support multiple
      key-value pairs separated by semicolons

    \b
    Example
    -------

    \b
    ```
    DTSTART: 2025-08-05 09:00
    DURATION: PT45M
    SUMMARY: Weekly Standup
    LOCATION: Zoom
    CATEGORIES: Work,Remote
    TIMEZONE: Europe/Lisbon
    PRIORITY: 3
    ALARM: -PT15M Reminder: Standup in 15 minutes
    RRULE: [[FREQ=WEEKLY;BYDAY=MO;COUNT=10]]
    DESCRIPTION: [[
    Weekly check-in with the team to discuss progress
    and blockers for the current sprint.
    ]]
    ```
    '''
    tz = context['config'].tz()
    # Read the event definition
    if edit and file is sys.stdin:
        # No file specified using an empty event
        tomorrow = datetime.now() + timedelta(days=1)
        event_string = EMPTY_EVENT % {
            'tz': f'{tz}',
            'start': tomorrow.replace(hour=9, minute=0, second=0).strftime('%Y-%m-%d %H:%M'),
            'end': tomorrow.replace(hour=9, minute=45, second=0).strftime('%Y-%m-%d %H:%M'),
            'cal': '\n'.join([('# ' + ln).strip() for ln in calmonth(tomorrow.year, tomorrow.month, 3)])
        }
    else:
        event_string = file.read()

    if edit:
        # Edit the file in $VISUAL or $EDITOR
        old_string = event_string
        event_string = edit_text_with_editor(event_string, suffix='.calendar_event')
        if old_string == event_string:
            click.confirm('No changes made. Do you want to continue?', abort=True)

    try:
        event_data = parse_event(event_string, tz)
    except ParserError as msg:
        raise click.UsageError(msg)

    event_ics = event_builder(event_data, tz)

    if dry_run:
        click.echo(event_ics)
    else:
        # Save the event to the calendar
        _, server = context['config'].get_server()
        calendar_id = context['config'].get_calendar()
        with dav.caldav_calendar(server, calendar_id) as calendar:
            # Check if the calendar supports events:
            if 'VEVENT' not in calendar.get_supported_components():
                raise click.UsageError('This calendar does not support events.')
            try:
                calendar.save_event(event_ics)
            except error.AuthorizationError as msg:
                raise click.UsageError(f'Error saving the event to the server: {msg}')

        click.echo('Event created successfully')


@click.command('delete', options_metavar='[options]')
@click.argument('event_id', metavar='<uid>')
@click.pass_obj
def delete_event(context, event_id):
    '''Delete an event on the server'''
    _, server = context['config'].get_server()
    calendar_id = context['config'].get_calendar()

    with dav.caldav_calendar_event(server, calendar_id, event_id) as event:
        event.delete()
    print('Event deleted')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime

import pytest

from caldavctl.object_parser import ParserError
from caldavctl.event_builder import RRuleValueParser, RRuleParser

rrule_test_data = (
    # Basic frequency tests
    ("FREQ=DAILY", {"freq": "DAILY"}),
    ("FREQ=WEEKLY", {"freq": "WEEKLY"}),
    ("FREQ=MONTHLY", {"freq": "MONTHLY"}),
    ("FREQ=YEARLY", {"freq": "YEARLY"}),
    ("FREQ=HOURLY", {"freq": "HOURLY"}),
    ("FREQ=MINUTELY", {"freq": "MINUTELY"}),
    ("FREQ=SECONDLY", {"freq": "SECONDLY"}),

    # Interval tests
    ("FREQ=DAILY;INTERVAL=1", {"freq": "DAILY", "interval": 1}),
    ("FREQ=WEEKLY;INTERVAL=2", {"freq": "WEEKLY", "interval": 2}),
    ("FREQ=MONTHLY;INTERVAL=3", {"freq": "MONTHLY", "interval": 3}),
    ("FREQ=YEARLY;INTERVAL=10", {"freq": "YEARLY", "interval": 10}),

    # Count tests
    ("FREQ=DAILY;COUNT=5", {"freq": "DAILY", "count": 5}),
    ("FREQ=WEEKLY;COUNT=10", {"freq": "WEEKLY", "count": 10}),

    # Until tests
    ("FREQ=DAILY;UNTIL=20250501T000000Z",
     {"freq": "DAILY",
      "until": datetime.datetime(2025, 5, 1, 0, 0, tzinfo=datetime.timezone.utc)}),
    ("FREQ=MONTHLY;UNTIL=20251231T235959Z",
     {"freq": "MONTHLY",
      "until": datetime.datetime(2025, 12, 31, 23, 59, 59, tzinfo=datetime.timezone.utc)}),

    # ByMonth tests
    ("FREQ=YEARLY;BYMONTH=1", {"freq": "YEARLY", "bymonth": [1]}),
    ("FREQ=YEARLY;BYMONTH=1,2,3", {"freq": "YEARLY", "bymonth": [1, 2, 3]}),
    ("FREQ=YEARLY;BYMONTH=1, 3, 5, 7, 9, 11", {"freq": "YEARLY", "bymonth": [1, 3, 5, 7, 9, 11]}),

    # ByMonthDay tests
    ("FREQ=MONTHLY;BYMONTHDAY=1", {"freq": "MONTHLY", "bymonthday": [1]}),
    ("FREQ=MONTHLY;BYMONTHDAY=1,15", {"freq": "MONTHLY", "bymonthday": [1, 15]}),
    ("FREQ=MONTHLY;BYMONTHDAY=-1", {"freq": "MONTHLY", "bymonthday": [-1]}),  # Last day of month
    ("FREQ=MONTHLY;BYMONTHDAY=1,-1", {"freq": "MONTHLY", "bymonthday": [1, -1]}),  # First and last day

    # ByDay tests
    ("FREQ=WEEKLY;BYDAY=MO", {"freq": "WEEKLY", "byday": ["MO"]}),
    ("FREQ=WEEKLY;BYDAY=MO,WE,FR", {"freq": "WEEKLY", "byday": ["MO", "WE", "FR"]}),
    ("FREQ=MONTHLY;BYDAY=1MO", {"freq": "MONTHLY", "byday": ["1MO"]}),  # First Monday
    ("FREQ=MONTHLY;BYDAY=-1FR", {"freq": "MONTHLY", "byday": ["-1FR"]}),  # Last Friday
    ("FREQ=MONTHLY;BYDAY=1MO,-1FR", {"freq": "MONTHLY", "byday": ["1MO", "-1FR"]}),  # First Monday and last Friday

    # ByYearDay tests
    ("FREQ=YEARLY;BYYEARDAY=1", {"freq": "YEARLY", "byyearday": [1]}),
    ("FREQ=YEARLY;BYYEARDAY=1,100,200", {"freq": "YEARLY", "byyearday": [1, 100, 200]}),
    ("FREQ=YEARLY;BYYEARDAY=-1", {"freq": "YEARLY", "byyearday": [-1]}),  # Last day of year
    ("FREQ=YEARLY;BYYEARDAY=1,-1", {"freq": "YEARLY", "byyearday": [1, -1]}),  # First and last day

    # ByWeekNo tests
    ("FREQ=YEARLY;BYWEEKNO=1", {"freq": "YEARLY", "byweekno": [1]}),
    ("FREQ=YEARLY;BYWEEKNO=1,10,20", {"freq": "YEARLY", "byweekno": [1, 10, 20]}),
    ("FREQ=YEARLY;BYWEEKNO=-1", {"freq": "YEARLY", "byweekno": [-1]}),  # Last week of year
    ("FREQ=YEARLY;BYWEEKNO=1,-1", {"freq": "YEARLY", "byweekno": [1, -1]}),  # First and last week

    # ByHour tests
    ("FREQ=DAILY;BYHOUR=9", {"freq": "DAILY", "byhour": [9]}),
    ("FREQ=DAILY;BYHOUR=9,12,17", {"freq": "DAILY", "byhour": [9, 12, 17]}),
    ("FREQ=DAILY;BYHOUR=10, 22", {"freq": "DAILY", "byhour": [10, 22]}),  # With space after comma

    # ByMinute tests
    ("FREQ=HOURLY;BYMINUTE=0", {"freq": "HOURLY", "byminute": [0]}),
    ("FREQ=HOURLY;BYMINUTE=0,15,30,45", {"freq": "HOURLY", "byminute": [0, 15, 30, 45]}),
    ("FREQ=HOURLY;BYMINUTE=0, 30", {"freq": "HOURLY", "byminute": [0, 30]}),  # With space after comma

    # BySecond tests
    ("FREQ=MINUTELY;BYSECOND=0", {"freq": "MINUTELY", "bysecond": [0]}),
    ("FREQ=MINUTELY;BYSECOND=0,15,30,45", {"freq": "MINUTELY", "bysecond": [0, 15, 30, 45]}),

    # BySetPos tests
    ("FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1",
     {"freq": "MONTHLY", "byday": ["MO", "TU", "WE", "TH", "FR"], "bysetpos": [1]}),
    ("FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1,-1",
     {"freq": "MONTHLY", "byday": ["MO", "TU", "WE", "TH", "FR"], "bysetpos": [1, -1]}),

    # WKST tests
    ("FREQ=WEEKLY;INTERVAL=1;WKST=SU", {"freq": "WEEKLY", "interval": 1, "wkst": "SU"}),
    ("FREQ=WEEKLY;INTERVAL=1;WKST=MO", {"freq": "WEEKLY", "interval": 1, "wkst": "MO"}),

    # Complex combinations
    ("FREQ=MONTHLY;BYDAY=MO,TU,WE,TH,FR;BYSETPOS=1;COUNT=3",
     {"freq": "MONTHLY", "byday": ["MO", "TU", "WE", "TH", "FR"], "bysetpos": [1], "count": 3}),
    ("FREQ=YEARLY;BYMONTH=1;BYMONTHDAY=1;UNTIL=20261231T235959Z",
     {"freq": "YEARLY",
      "bymonth": [1],
      "bymonthday": [1],
      "until": datetime.datetime(2026, 12, 31, 23, 59, 59, tzinfo=datetime.timezone.utc)
      }),
    ("FREQ=WEEKLY;INTERVAL=2;BYDAY=MO,WE,FR;WKST=SU;COUNT=10",
     {"freq": "WEEKLY", "interval": 2, "byday": ["MO", "WE", "FR"], "wkst": "SU", "count": 10}),
    ("FREQ=DAILY;INTERVAL=1;BYHOUR=9,17;BYMINUTE=0,30",
     {"freq": "DAILY", "interval": 1, "byhour": [9, 17], "byminute": [0, 30]}),

    # Case insensitivity tests
    ("freq=daily", {"freq": "DAILY"}),
    ("Freq=Daily;Interval=1", {"freq": "DAILY", "interval": 1}),

    # Whitespace handling
    ("FREQ=DAILY; INTERVAL=1", {"freq": "DAILY", "interval": 1}),
    ("FREQ=DAILY ; INTERVAL = 1", {"freq": "DAILY", "interval": 1}),

    # Empty values (if applicable)
    ("FREQ=DAILY;BYDAY=", {"freq": "DAILY"}),

    # Multiple semicolons
    (";;FREQ=MINUTELY;", {"freq": "MINUTELY"}),
    ("FREQ=DAILY;;INTERVAL=1", {"freq": "DAILY", "interval": 1}),

    # Trailing semicolon
    ("FREQ=DAILY;", {"freq": "DAILY"}),
    # Leading semicolon
    (";FREQ=DAILY", {"freq": "DAILY"}),
)


@pytest.mark.parametrize('sample, result', rrule_test_data)
def test_rrule_data(sample, result):
    print(f'\nTesting RRULE: {sample}')

    LEXICON = RRuleValueParser.LEXICON
    MANDATORY_KEYS = RRuleValueParser.MANDATORY_KEYS
    tokens = RRuleParser(sample, LEXICON, MANDATORY_KEYS).run()

    print('Computed tokens:', tokens)
    assert tokens == result


rrule_error_test_data = (
    # Invalid frequency value
    ('freq=DIÁRIO', 'Error evaluating key "FREQ": Element DIÁRIO not in allowed values'),

    # Missing mandatory FREQ key
    ('COUNT=5', 'Missing mandatory key: "freq"'),

    # Invalid until date
    ('FREQ=daily;UNTIL=2025FEB',
     'Error evaluating key "UNTIL": Invalid date format for: "2025FEB"'),

    # Invalid day name
    ("FREQ=WEEKLY;BYDAY=FU", 'Error evaluating key "BYDAY": Invalid week day "FU"'),

    # Invalid month number
    ("FREQ=YEARLY;BYMONTH=13",
     'Error evaluating key "BYMONTH": the value is greater than the maximum value'),

    # Both COUNT and UNTIL (should be exclusive)
    ("FREQ=DAILY;COUNT=5;UNTIL=20250501T000000Z",
     'Only one of these keys can be present: count, until'),

    # Invalid INTERVAL (negative value)
    ("FREQ=DAILY;INTERVAL=-1",
     'Error evaluating key "INTERVAL": the value is less than the minimum value'),

    # Multiple equal signs
    ("FREQ==DAILY", 'Error evaluating key "FREQ": Element =DAILY not in allowed values'),
)


@pytest.mark.parametrize('sample, error_message', rrule_error_test_data)
def test_rrule_errors(sample, error_message):
    print(f'\nTesting RRULE: {sample}')
    LEXICON = RRuleValueParser.LEXICON
    MANDATORY_KEYS = RRuleValueParser.MANDATORY_KEYS
    EXLUSIVE_KEYS = RRuleValueParser.EXCLUSIVE_KEYS

    with pytest.raises(ParserError, match=error_message):
        RRuleParser(sample, LEXICON, MANDATORY_KEYS, EXLUSIVE_KEYS).run()



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later


from caldavctl.click_utils import OptionsCompatibility

from click.testing import CliRunner

from caldavctl import cli


def test_click_utils_compatibility():
    compatibility = OptionsCompatibility(['today', 'week', 'dtstart', 'dtend'])
    compatibility.set_exception(['dtstart', 'dtend'])

    test_compatibility = {
        'today': {
            'week': False,
            'dtstart': False,
            'dtend': False,
        },
        'week': {
            'today': False,
            'dtstart': False,
            'dtend': False,
        },
        'dtstart': {
            'today': False,
            'week': False,
            'dtend': True,
        },
        'dtend': {
            'today': False,
            'week': False,
            'dtstart': True,
        },
    }

    assert test_compatibility == compatibility.compatibility


def test_click_utils_incompatible():
    '''
    Test incompatible options
    '''
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ['event', 'list', '-s', '2025-01-01', '--week', '1']
    )
    assert 'Error, the options "dtstart" and "week" can\'t be used together.' in result.output


def test_click_utils_incompatible_optional_value():
    '''
    Test incompatible options when one of the options has optional values using
    the default value.
    '''
    runner = CliRunner()
    result = runner.invoke(
        cli.cli_group,
        ['event', 'list', '-s', '2025-01-01', '--week']
    )
    assert 'Error, the options "dtstart" and "week" can\'t be used together.' in result.output



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import caldav
import click
from caldav.lib import error

from caldavctl import dav, get_name
from caldavctl.templates import render
from caldavctl.utils import (
    ESCAPE_SEQUENCES,
)


# CALENDARS

@click.command('list', options_metavar='[options]')
@click.option('-tf', '--template-file', default='calendar.txt',
              help='Template used to format the output', metavar='<file>')
@click.option('--json',
              is_flag=True, show_default=True, default=False,
              help='Output JSON instead of using a template')
@click.pass_obj
def list_calendars(context, template_file, json):
    '''
    List the available calendars present on the server or servers

    caldavctl uses templates to display the list command output by default. If
    not specified otherwise it uses the `calendar.txt` template. The used
    template can be specified using the -tf/--template-file option. First we
    check the current directory for the existence of the template file. If it's
    not found we check the `<share>/caldavctl/templates` directory.

    Currently we have the following templates:

    \b
        * calendar.txt - output with colors (default)
        * calendar-nocolor.txt
        * calendar.html

    If the --json flag is used, the list command will output JSON. This output
    does not go through the template system, and any template options will be
    ignored when using --json.
    '''
    calendar_list = []
    for name, server in context['config'].servers():
        with caldav.DAVClient(**server) as client:
            for calendar in client.principal().calendars():
                calendar_list.append({
                    'server': name,
                    'name': calendar.name,
                    'id': calendar.id,
                    'components': calendar.get_supported_components(),
                    'url': f'{calendar.url}',
                })

    if json:
        import json
        click.echo(json.dumps(calendar_list))
        return

    # Template output
    template_context = {
        'calendar_list': calendar_list,
    }
    template_context.update(ESCAPE_SEQUENCES)

    txt = render(get_name(), template_file, template_context)
    click.echo(txt.strip())


@click.command('create', options_metavar='[options]')
@click.argument('name', metavar='<calendar>')
@click.option('--cal-id', help='Calendar UID to use in the new calendar', metavar='<uid>')
@click.pass_obj
def create_calendar(context, name, cal_id=None):
    '''
    Create a calendar on the default server or optionally in another server

    <calendar> - calendar name
    '''
    _, server = context['config'].get_server()

    with caldav.DAVClient(**server) as client:
        principal = client.principal()
        try:
            new_calendar = principal.make_calendar(name=name, cal_id=cal_id)
        except error.AuthorizationError as msg:
            raise click.UsageError(f'Error creating the calendar (maybe duplicate UID?) with: {msg}')

        print(f'Calendar "{name}" created.')
    return new_calendar


@click.command('delete', options_metavar='[options]')
@click.argument('calendar-id', metavar='<uid>')
@click.pass_obj
def delete_calendar(context, calendar_id):
    '''
    Delete a calendar from the default server or optionally from another
    server. It's possible to have calendars with the same name, so we use the
    id to identify the calendar to delete.
    '''
    _, server = context['config'].get_server()

    with dav.caldav_calendar(server, calendar_id) as calendar:
        name = calendar.name
        calendar.delete()
        print(f'Calendar "{name}" deleted.')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import uuid
from datetime import datetime

from icalendar import Alarm, Calendar, Event, vRecur

from caldavctl import get_name, get_version
from caldavctl.object_parser import (
    DatetimeParser,
    EventParser,
    ParserError,
    RRuleParser,
    TimezoneParser,
    ValueParser,
    choices_parser,
    integer_list_parser,
    integer_parser,
    list_parser,
    string_parser,
)
from caldavctl.utils import check_if_naive, duration_to_timedelta, duration_validation


class AlarmParser(ValueParser):
    def tokenize(self):
        self.tokens = self.txt.split(maxsplit=1)

    def validate(self):
        duration = self.tokens[0].upper().strip()
        if not duration_validation(duration):
            raise ParserError(f'Invalid duration "{self.tokens[0]}" for alarm trigger.')
        self.tokens[0] = duration_to_timedelta(duration)

        alarm = Alarm()
        alarm.add('trigger', self.tokens[0])
        if len(self.tokens) == 2:
            alarm.add('description', self.tokens[1])
        else:
            alarm.add('description', 'Default caldavctl description')
        alarm.add('action', 'display')

        self.value = alarm


class DurationParser(ValueParser):
    def tokenize(self):
        self.tokens = self.txt

    def validate(self):
        duration = self.tokens.upper().strip()
        if not duration_validation(duration):
            raise ParserError(f'Invalid duration "{self.tokens}".')
        self.value = duration_to_timedelta(duration)


class ByDayParser(ValueParser):
    WEEKDAYS = ['SU', 'MO', 'TU', 'WE', 'TH', 'FR', 'SA']

    def tokenize(self):
        self.tokens = [el.strip() for el in self.txt.split(',')]

    def validate(self):
        for wday in self.tokens:
            week_day = wday[-2:].upper()
            if week_day not in self.WEEKDAYS:
                raise ParserError(f'Invalid week day "{week_day}"')
            ordwk = wday[:-2]
            if ordwk:
                try:
                    ordwk = int(ordwk)
                except ValueError:
                    raise ParserError(f'Invalid week number "{ordwk}"')
                if abs(ordwk) > 53 or ordwk == 0:
                    raise ParserError(f'Out of range week number "{ordwk}"')

        self.value = self.tokens


class RRuleValueParser(ValueParser):
    LEXICON = {
        'freq': choices_parser(['SECONDLY', 'MINUTELY', 'HOURLY', 'DAILY',
                                'WEEKLY', 'MONTHLY', 'YEARLY']),
        'until': DatetimeParser,
        'count': integer_parser(min=1),
        'interval': integer_parser(min=1),
        'bysecond': integer_list_parser(0, 60),
        'byminute': integer_list_parser(0, 59),
        'byhour': integer_list_parser(0, 23),
        'byday': ByDayParser,
        'bymonthday': integer_list_parser(-31, 31, exception=[0]),
        'byyearday': integer_list_parser(-366, 366, exception=[0]),
        'byweekno': integer_list_parser(-53, 53, exception=[0]),
        'bymonth': integer_list_parser(1, 12),
        'bysetpos': integer_list_parser(-366, 366, exception=[0]),
        'wkst': choices_parser(['SU', 'MO', 'TU', 'WE', 'TH', 'FR', 'SA'])
    }

    MANDATORY_KEYS = ('freq',)

    EXCLUSIVE_KEYS = (['count', 'until'],)  # Only one of this keys may be present

    def tokenize(self):
        self.tokens = RRuleParser(
            self.txt,
            self.LEXICON,
            self.MANDATORY_KEYS,
            self.EXCLUSIVE_KEYS).run()

    def validate(self):
        self.value = vRecur(self.tokens)


EVENT_LEXICON = {
    # token: type
    'summary': string_parser(),
    'location': string_parser(),
    'categories': list_parser(),
    'dtstart': DatetimeParser,
    'dtend': DatetimeParser,
    'duration': DurationParser,
    'description': string_parser(),
    'percentage': integer_parser(min=0, max=100),
    'timezone': TimezoneParser,
    'priority': integer_parser(min=1, max=9),
    'alarm': AlarmParser,
    'rrule': RRuleValueParser,
}

OPTIONAL_KEYS = (
    'summary',
    'location',
    'categories',
    'description',
    'percentage',
    'timezone',
    'priority',
    'alarm',
    'rrule',
)

MANDATORY_KEYS = (
    'dtstart',
    ['dtend', 'duration'],  # One of these key must exist
)


def parse_event(event_icalendar, timezone):
    result = EventParser(event_icalendar, EVENT_LEXICON, MANDATORY_KEYS).run()

    # Check if the start and/or end dates are naïve
    tz = timezone if 'timezone' not in result else result['timezone']
    result['dtstart'] = check_if_naive(result['dtstart'], tz)
    if 'dtend' in result:
        result['dtend'] = check_if_naive(result['dtend'], tz)

    return result


def event_builder(event_data, tz):
    calendar = Calendar()
    calendar.add('version', '2.0')
    calendar.add('prodid', f'-//NA//{get_name()} V{get_version()}//EN')

    event = Event()
    # Mandatory keys in a VEVENT object
    event.add('uid', str(uuid.uuid4()))
    event.add('dtstamp', datetime.now(tz))
    event.add('dtstart', event_data['dtstart'])
    if 'dtend' in event_data:
        event.add('dtend', event_data['dtend'])
    elif 'duration' in event_data:
        event.add('duration', event_data['duration'])
    # Optional keys:
    for key in OPTIONAL_KEYS:
        if key in event_data:
            if isinstance(event_data[key], list):
                for value in event_data[key]:
                    if key == 'alarm':
                        event.add_component(value)
                    else:
                        event.add(key, value)
            else:
                if key == 'alarm':
                    event.add_component(event_data[key])
                else:
                    event.add(key, event_data[key])

    calendar.add_component(event)

    return calendar.to_ical().decode('utf-8')



# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytest
from unittest.mock import MagicMock
from click.testing import CliRunner
from caldavctl.event import create_event
from zoneinfo import ZoneInfo


@pytest.fixture
def mock_context():
    """Fixture for the CLI context."""
    return {
        "config": MagicMock(
            servers=MagicMock(return_value=[("default", {"url": "testserver.com"})]),
            get_server=MagicMock(return_value=("default", {"url": "testserver.com"})),
            tz=MagicMock(return_value=ZoneInfo("Europe/London"))
        )
    }


def test_create_event_one_alarm(mock_context, tmp_path, monkeypatch):
    """Test listing calendars."""
    event_data = """
        SUMMARY: This is a description
        LOCATION: [[Timbuktu]] # With a comment
        CATEGORIES: Birthday, Shopping, Training
        TIMEZONE: Europe/Madrid
        DTSTART: 2024-01-20 09:00
        DTEND: 2024-01-20 09:45
        ALARM: P0D One alarm
        DESCRIPTION: [[
        Once upon a time,
        There was a little lamb
        ]]
        """
    # Create a temporary file
    temp_file = tmp_path / "event.txt"
    temp_file.write_text(event_data)

    runner = CliRunner()
    result = runner.invoke(create_event, ['--dry-run', "--file", str(temp_file)], obj=mock_context)

    assert result.exit_code == 0
    assert "BEGIN:VALARM\nACTION:display\nDESCRIPTION:One alarm\nTRIGGER:P0D" in result.output


def test_create_event_two_alarms(mock_context, tmp_path, monkeypatch):
    """Test listing calendars."""
    event_data = """
        SUMMARY: This is a description
        LOCATION: [[Timbuktu]] # With a comment
        CATEGORIES: Birthday, Shopping, Training
        TIMEZONE: Europe/Madrid
        DTSTART: 2024-01-20 09:00
        DTEND: 2024-01-20 09:45
        ALARM: P0D One alarm
        ALARM: -PT1H Another alarm
        DESCRIPTION: [[
        Once upon a time,
        There was a little lamb
        ]]
        """
    # Create a temporary file
    temp_file = tmp_path / "event.txt"
    temp_file.write_text(event_data)

    runner = CliRunner()
    result = runner.invoke(create_event, ['--dry-run', "--file", str(temp_file)], obj=mock_context)

    assert result.exit_code == 0
    assert "BEGIN:VALARM\nACTION:display\nDESCRIPTION:One alarm\nTRIGGER:P0D" in result.output
    assert "BEGIN:VALARM\nACTION:display\nDESCRIPTION:Another alarm\nTRIGGER:-PT1H\nEND:VALARM" in result.output




# SPDX-FileCopyrightText: 2024-present Helder Guerreiro <helder@paxjulia.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

'''
Utilities
'''

import zoneinfo
import pprint

import click

from caldavctl.templates import share_directory_list
from caldavctl import get_name


@click.command('list-timezones', options_metavar='[options]')
@click.option('-s', '--search',
              default=None, metavar='<query>', type=str,
              help='Search time zone')
@click.pass_obj
def list_timezones(context, search):
    '''
    List available time zones

    This command display the available time zones in python's zoneinfo module.
    Use the -s/--search option to search for a specific time zone.
    '''
    timezones = zoneinfo.available_timezones()

    if search:
        result = []
        for tz in timezones:
            if search.lower() in tz.lower():
                result.append(tz)
        if not result:
            result.append(f'No time zone "{search}" found.')
    else:
        result = list(timezones)

    result.sort()

    for tz in result:
        click.echo(tz)


@click.command('config-file')
@click.pass_obj
def config_file(context):
    '''
    Show the paths where the configuration files might reside
    '''
    click.echo(f'The configuration file used is:\n{context['config'].config_file}')


@click.command('current-config')
@click.pass_obj
def current_config(context):
    '''
    Current configuration
    '''
    click.echo('The current configuration is:')
    pprint.pprint(context['config'].config)


@click.command('template-directories')
@click.pass_obj
def template_directories(context):
    '''
    Show the paths from where the templates are read

    When searching for templates we first test for the existence of the
    template in the current directory. If the template is not found the
    template file is searched on the directories listed by this command.

    The existing directories are marked with an asterisk (*).
    '''
    app_name = get_name()
    click.echo('Besides the current durectory, the following paths are\n'
               'considered when searching for templates:\n')
    for path in share_directory_list(app_name):
        p = path / 'templates'
        click.echo(f'{p}{' *' if p.exists() else ''}')
